@@ -1,6 +1,6 @@
-/* 
+/*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2011 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -33,8 +33,9 @@
 #include <unistd.h>
 
 #include "arp.h"
-#include "bind.h"
+#include "ipv4.h"
 #include "common.h"
+#include "dhcp.h"
 #include "dhcpcd.h"
 #include "eloop.h"
 #include "if-options.h"
@@ -45,80 +46,92 @@
 	(sizeof(struct arphdr) + (2 * sizeof(uint32_t)) + (2 * HWADDR_LEN))
 
 static int
-send_arp(const struct interface *iface, int op, in_addr_t sip, in_addr_t tip)
+arp_send(const struct interface *ifp, int op, in_addr_t sip, in_addr_t tip)
 {
 	uint8_t arp_buffer[ARP_LEN];
 	struct arphdr ar;
 	size_t len;
 	uint8_t *p;
-	int retval;
 
-	ar.ar_hrd = htons(iface->family);
+	ar.ar_hrd = htons(ifp->family);
 	ar.ar_pro = htons(ETHERTYPE_IP);
-	ar.ar_hln = iface->hwlen;
+	ar.ar_hln = ifp->hwlen;
 	ar.ar_pln = sizeof(sip);
 	ar.ar_op = htons(op);
-	memcpy(arp_buffer, &ar, sizeof(ar));
-	p = arp_buffer + sizeof(ar);
-	memcpy(p, iface->hwaddr, iface->hwlen);
-	p += iface->hwlen;
-	memcpy(p, &sip, sizeof(sip));
-	p += sizeof(sip);
-	/* ARP requests should ignore this */
-	retval = iface->hwlen;
-	while (retval--)
-		*p++ = '\0';
-	memcpy(p, &tip, sizeof(tip));
-	p += sizeof(tip);
+
+	p = arp_buffer;
+	len = sizeof(arp_buffer);
+
+#define CHECK(fun, b, l)						\
+	do {								\
+		if (len < (l))						\
+			goto eexit;					\
+		fun(p, (b), (l));					\
+		p += (l);						\
+		len -= (l);						\
+	} while (/* CONSTCOND */ 0)
+#define APPEND(b, l)	CHECK(memcpy, b, l)
+#define ZERO(l)		CHECK(memset, 0, l)
+
+	APPEND(&ar, sizeof(ar));
+	APPEND(ifp->hwaddr, ifp->hwlen);
+	APPEND(&sip, sizeof(sip));
+	ZERO(ifp->hwlen);
+	APPEND(&tip, sizeof(tip));
 	len = p - arp_buffer;
-	retval = send_raw_packet(iface, ETHERTYPE_ARP, arp_buffer, len);
-	return retval;
+	return ipv4_sendrawpacket(ifp, ETHERTYPE_ARP, arp_buffer, len);
+
+eexit:
+	errno = ENOBUFS;
+	return -1;
 }
 
 static void
-handle_arp_failure(struct interface *iface)
+arp_failure(struct interface *ifp)
 {
+	const struct dhcp_state *state = D_CSTATE(ifp);
 
 	/* If we failed without a magic cookie then we need to try
 	 * and defend our IPv4LL address. */
-	if ((iface->state->offer != NULL &&
-		iface->state->offer->cookie != htonl(MAGIC_COOKIE)) ||
-	    (iface->state->new != NULL &&
-		iface->state->new->cookie != htonl(MAGIC_COOKIE)))
+	if ((state->offer != NULL &&
+		state->offer->cookie != htonl(MAGIC_COOKIE)) ||
+	    (state->new != NULL &&
+		state->new->cookie != htonl(MAGIC_COOKIE)))
 	{
-		handle_ipv4ll_failure(iface);
+		ipv4ll_handle_failure(ifp);
 		return;
 	}
 
-	unlink(iface->leasefile);
-	if (!iface->state->lease.frominfo)
-		send_decline(iface);
-	close_sockets(iface);
-	delete_timeout(NULL, iface);
-	if (iface->state->lease.frominfo)
-		start_interface(iface);
+	unlink(state->leasefile);
+	if (!state->lease.frominfo)
+		dhcp_decline(ifp);
+	dhcp_close(ifp);
+	eloop_timeout_delete(NULL, ifp);
+	if (state->lease.frominfo)
+		start_interface(ifp);
 	else
-		add_timeout_sec(DHCP_ARP_FAIL, start_interface, iface);
+		eloop_timeout_add_sec(DHCP_ARP_FAIL, start_interface, ifp);
 }
 
 static void
-handle_arp_packet(void *arg)
+arp_packet(void *arg)
 {
-	struct interface *iface = arg;
+	struct interface *ifp = arg;
 	uint8_t arp_buffer[ARP_LEN];
 	struct arphdr ar;
 	uint32_t reply_s;
 	uint32_t reply_t;
 	uint8_t *hw_s, *hw_t;
 	ssize_t bytes;
-	struct if_state *state = iface->state;
-	struct if_options *opts = state->options;
+	struct dhcp_state *state;
+	struct if_options *opts = ifp->options;
 	const char *hwaddr;
 	struct in_addr ina;
 
+	state = D_STATE(ifp);
 	state->fail.s_addr = 0;
 	for(;;) {
-		bytes = get_raw_packet(iface, ETHERTYPE_ARP,
+		bytes = ipv4_getrawpacket(ifp, ETHERTYPE_ARP,
 		    arp_buffer, sizeof(arp_buffer), NULL);
 		if (bytes == 0 || bytes == -1)
 			return;
@@ -143,8 +156,8 @@ handle_arp_packet(void *arg)
 		if ((hw_t + ar.ar_hln + ar.ar_pln) - arp_buffer > bytes)
 			continue;
 		/* Ignore messages from ourself */
-		if (ar.ar_hln == iface->hwlen &&
-		    memcmp(hw_s, iface->hwaddr, iface->hwlen) == 0)
+		if (ar.ar_hln == ifp->hwlen &&
+		    memcmp(hw_s, ifp->hwaddr, ifp->hwlen) == 0)
 			continue;
 		/* Copy out the IP addresses */
 		memcpy(&reply_s, hw_s + ar.ar_hln, ar.ar_pln);
@@ -162,13 +175,13 @@ handle_arp_packet(void *arg)
 			    (size_t)ar.ar_hln);
 			syslog(LOG_INFO,
 			    "%s: found %s on hardware address %s",
-			    iface->name, inet_ntoa(ina), hwaddr);
-			if (select_profile(iface, hwaddr) == -1 &&
+			    ifp->name, inet_ntoa(ina), hwaddr);
+			if (select_profile(ifp, hwaddr) == -1 &&
 			    errno == ENOENT)
-				select_profile(iface, inet_ntoa(ina));
-			close_sockets(iface);
-			delete_timeout(NULL, iface);
-			start_interface(iface);
+				select_profile(ifp, inet_ntoa(ina));
+			dhcp_close(ifp);
+			eloop_timeout_delete(NULL, ifp);
+			start_interface(ifp);
 			return;
 		}
 
@@ -179,49 +192,51 @@ handle_arp_packet(void *arg)
 			state->fail.s_addr = state->offer->yiaddr;
 
 		/* Handle IPv4LL conflicts */
-		if (IN_LINKLOCAL(htonl(iface->addr.s_addr)) &&
-		    (reply_s == iface->addr.s_addr ||
-			(reply_s == 0 && reply_t == iface->addr.s_addr)))
-			state->fail.s_addr = iface->addr.s_addr;
+		if (IN_LINKLOCAL(htonl(state->addr.s_addr)) &&
+		    (reply_s == state->addr.s_addr ||
+			(reply_s == 0 && reply_t == state->addr.s_addr)))
+			state->fail.s_addr = state->addr.s_addr;
 
 		if (state->fail.s_addr) {
 			syslog(LOG_ERR, "%s: hardware address %s claims %s",
-			    iface->name,
+			    ifp->name,
 			    hwaddr_ntoa((unsigned char *)hw_s,
 				(size_t)ar.ar_hln),
 			    inet_ntoa(state->fail));
 			errno = EEXIST;
-			handle_arp_failure(iface);
+			arp_failure(ifp);
 			return;
 		}
 	}
 }
 
 void
-send_arp_announce(void *arg)
+arp_announce(void *arg)
 {
-	struct interface *iface = arg;
-	struct if_state *state = iface->state;
+	struct interface *ifp = arg;
+	struct dhcp_state *state = D_STATE(ifp);
 	struct timeval tv;
 
-	if (iface->arp_fd == -1) {
-		open_socket(iface, ETHERTYPE_ARP);
-		add_event(iface->arp_fd, handle_arp_packet, iface);
+	if (state->new == NULL)
+		return;
+	if (state->arp_fd == -1) {
+		ipv4_opensocket(ifp, ETHERTYPE_ARP);
+		eloop_event_add(state->arp_fd, arp_packet, ifp);
 	}
-	if (++state->claims < ANNOUNCE_NUM)	
+	if (++state->claims < ANNOUNCE_NUM)
 		syslog(LOG_DEBUG,
 		    "%s: sending ARP announce (%d of %d), "
 		    "next in %d.00 seconds",
-		    iface->name, state->claims, ANNOUNCE_NUM, ANNOUNCE_WAIT);
+		    ifp->name, state->claims, ANNOUNCE_NUM, ANNOUNCE_WAIT);
 	else
 		syslog(LOG_DEBUG,
 		    "%s: sending ARP announce (%d of %d)",
-		    iface->name, state->claims, ANNOUNCE_NUM);
-	if (send_arp(iface, ARPOP_REQUEST,
+		    ifp->name, state->claims, ANNOUNCE_NUM);
+	if (arp_send(ifp, ARPOP_REQUEST,
 		state->new->yiaddr, state->new->yiaddr) == -1)
 		syslog(LOG_ERR, "send_arp: %m");
 	if (state->claims < ANNOUNCE_NUM) {
-		add_timeout_sec(ANNOUNCE_WAIT, send_arp_announce, iface);
+		eloop_timeout_add_sec(ANNOUNCE_WAIT, arp_announce, ifp);
 		return;
 	}
 	if (state->new->cookie != htonl(MAGIC_COOKIE)) {
@@ -234,25 +249,31 @@ send_arp_announce(void *arg)
 		tv.tv_sec = state->interval - DHCP_RAND_MIN;
 		tv.tv_usec = arc4random() % (DHCP_RAND_MAX_U - DHCP_RAND_MIN_U);
 		timernorm(&tv);
-		add_timeout_tv(&tv, start_discover, iface);
+		eloop_timeout_add_tv(&tv, dhcp_discover, ifp);
 	} else {
-		delete_event(iface->arp_fd);
-		close(iface->arp_fd);
-		iface->arp_fd = -1;
+		eloop_event_delete(state->arp_fd);
+		close(state->arp_fd);
+		state->arp_fd = -1;
 	}
 }
 
 void
-send_arp_probe(void *arg)
+arp_probe(void *arg)
 {
-	struct interface *iface = arg;
-	struct if_state *state = iface->state;
+	struct interface *ifp = arg;
+	struct dhcp_state *state = D_STATE(ifp);
 	struct in_addr addr;
 	struct timeval tv;
 	int arping = 0;
 
-	if (state->arping_index < state->options->arping_len) {
-		addr.s_addr = state->options->arping[state->arping_index];
+	if (state->arp_fd == -1) {
+		if (ipv4_opensocket(ifp, ETHERTYPE_ARP) == -1)
+			return;
+		eloop_event_add(state->arp_fd, arp_packet, ifp);
+	}
+
+	if (state->arping_index < ifp->options->arping_len) {
+		addr.s_addr = ifp->options->arping[state->arping_index];
 		arping = 1;
 	} else if (state->offer) {
 		if (state->offer->yiaddr)
@@ -260,49 +281,47 @@ send_arp_probe(void *arg)
 		else
 			addr.s_addr = state->offer->ciaddr;
 	} else
-		addr.s_addr = iface->addr.s_addr;
+		addr.s_addr = state->addr.s_addr;
 
-	if (iface->arp_fd == -1) {
-		open_socket(iface, ETHERTYPE_ARP);
-		add_event(iface->arp_fd, handle_arp_packet, iface);
-	}
 	if (state->probes == 0) {
 		if (arping)
-			syslog(LOG_INFO, "%s: searching for %s",
-			    iface->name, inet_ntoa(addr));
+			syslog(LOG_DEBUG, "%s: searching for %s",
+			    ifp->name, inet_ntoa(addr));
 		else
-			syslog(LOG_INFO, "%s: checking for %s",
-			    iface->name, inet_ntoa(addr));
+			syslog(LOG_DEBUG, "%s: checking for %s",
+			    ifp->name, inet_ntoa(addr));
 	}
 	if (++state->probes < PROBE_NUM) {
 		tv.tv_sec = PROBE_MIN;
 		tv.tv_usec = arc4random() % (PROBE_MAX_U - PROBE_MIN_U);
 		timernorm(&tv);
-		add_timeout_tv(&tv, send_arp_probe, iface);
+		eloop_timeout_add_tv(&tv, arp_probe, ifp);
 	} else {
 		tv.tv_sec = ANNOUNCE_WAIT;
 		tv.tv_usec = 0;
 		if (arping) {
 			state->probes = 0;
-			if (++state->arping_index < state->options->arping_len)
-				add_timeout_tv(&tv, send_arp_probe, iface);
+			if (++state->arping_index < ifp->options->arping_len)
+				eloop_timeout_add_tv(&tv, arp_probe, ifp);
 			else
-				add_timeout_tv(&tv, start_interface, iface);
+				eloop_timeout_add_tv(&tv, start_interface, ifp);
 		} else
-			add_timeout_tv(&tv, bind_interface, iface);
+			eloop_timeout_add_tv(&tv, dhcp_bind, ifp);
 	}
 	syslog(LOG_DEBUG,
 	    "%s: sending ARP probe (%d of %d), next in %0.2f seconds",
-	    iface->name, state->probes ? state->probes : PROBE_NUM, PROBE_NUM,
+	    ifp->name, state->probes ? state->probes : PROBE_NUM, PROBE_NUM,
 	    timeval_to_double(&tv));
-	if (send_arp(iface, ARPOP_REQUEST, 0, addr.s_addr) == -1)
+	if (arp_send(ifp, ARPOP_REQUEST, 0, addr.s_addr) == -1)
 		syslog(LOG_ERR, "send_arp: %m");
 }
 
 void
-start_arping(struct interface *iface)
+arp_start(struct interface *ifp)
 {
-	iface->state->probes = 0;
-	iface->state->arping_index = 0;
-	send_arp_probe(iface);
+	struct dhcp_state *state = D_STATE(ifp);
+
+	state->probes = 0;
+	state->arping_index = 0;
+	arp_probe(ifp);
 }
